# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Plugin module to find whether a large discharge is required for an ADM centrifuge """


__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2021 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import os
import gc
import copy
import logging

from foglamp.common import iprpc, logger
import filter_ingest

from foglamp.plugins.filter.ADM_LD_prediction.model_inference_handler import ModelInferenceHandler

_LOGGER = logger.setup(__name__, level=logging.INFO)  # WARN

_ADM_LD_NEEDED_ASSET = "centrifugeHealth"
_VIBRATION_TIME_SERIES_ASSET = "vibration"

KEY_DATA1 = 'Channel0'  # translate to canonical names
KEY_DATA2 = 'Channel1'
KEY_DATA3 = 'Channel2'
KEY_DATA4 = 'Channel3'
KEY_TS = 'user_ts'


_DEFAULT_CONFIG = {
    'plugin': {   # mandatory filter
        'description': 'filter plugin to detect need for large discharge',
        'type': 'string',
        'default': 'ADM_LD_prediction',
        'readonly': 'true'
    },
    'enable': {    # recommended filter
        'description': 'Enable LD prediction plugin',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Enabled',
        'order': "1"
    },
    'logLevel': {
        'description': 'log level for diagnostic or error output',
        'type': 'enumeration',
        'options': ['ERROR', 'WARN', 'INFO', 'DEBUG'],
        'default': 'INFO',
        'displayName': 'Log level',
        'order': "2"
    },
    'timeseriesAsset': {
        'description': 'Name of asset that has the vibration time series',
        'type': 'string',
        'default': _VIBRATION_TIME_SERIES_ASSET,
        'displayName': 'Asset from which vibration data is read',
        'order': "3"
    },
    'centrifugeHealthAsset': {
        'description': 'Name of asset that indicates centrifuge health',
        'type': 'string',
        'default': _ADM_LD_NEEDED_ASSET,
        'displayName': 'Output asset indicating centrifuge health',
        'order': "4"
    },
    'useCols': {
        'description': 'Comma separated names of ts and data channels',
        'type': 'string',
        'default': 'user_ts,Channel1,Channel2',
        'displayName': 'Comma separated TS and Data channel(s)',
        'order': "5"
    },
    'samplingRate': {
        'description': 'Readings per sec',
        'type': 'integer',
        'default': '8000',
        'displayName': 'Sampling rate (per sec)',
        'order': "6"
    },
    'modelDirPath': {
        'description': 'Model dir path',
        'type': 'string',
        'default': os.path.join(os.path.dirname(__file__), "model"),
        'displayName': 'Model directory path',
        'order': "7"
    },
    'modelType': {
        'description': 'Model type to use for predicting need for large discharge',
        'type': 'enumeration',
        'default': 'Autoencoder',
        'options': ['Autoencoder'],
        'displayName': 'Model type',
        'order': '8'
    },
    'enableTpu': {
        'description': 'Connect the Coral Edge TPU and enable this',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Enable Edge TPU',
        'order': '9',
        'validity': "modelType == 'Autoencoder'",
        'readonly': 'true'
    },
    'forwardData': {
        'description': 'Forward received data down the filter chain',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Forward data',
        'order': "10"
    },
    'evaluationInterval': {
        'description': 'Time interval after which to evaluate vibration data through the model '
                       '(time interval strings must be supported by pandas.to_timedelta API '
                       'e.g. 15s, 30m, 2hr, 1day 4hr 30m 10s)',
        'type': 'string',
        'default': '30min',
        'displayName': 'Vibration data evaluation interval',
        'order': "11"
    },
    'numEvaluations': {
        'description': 'Number of consecutive evaluations to consider at above specified interval',
        'type': 'integer',
        'default': '10',
        'displayName': 'Number of consecutive evaluations to consider (set)',
        'order': "12"
    },
    'numAnomalies': {
        'description': 'Number of anomalous evaluations in total number of evaluations set above',
        'type': 'integer',
        'default': '8',
        'displayName': 'Number of anomalous evaluations in set to declare anomaly',
        'order': "13"
    }
}


def plugin_info():
    """ Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    _LOGGER.info("plugin_info called")
    return {
        'name': 'ADM_LD_prediction',
        'version': '2.0.0',
        'mode': 'none',
        'type': 'filter',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


LOGLEVEL = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARN': logging.WARN,
    'ERROR': logging.ERROR
}


class CentrifugeState:
    def __init__(self, config, ingest_ref, callback):
        self.config = config
        self.ingest_ref = ingest_ref
        self.callback = callback

        self.health_asset = config['centrifugeHealthAsset']['value']
        self.ts_asset = config['timeseriesAsset']['value']
        self.forwardData = (config['forwardData']['value'] == 'true')
        _LOGGER.info("self.forwardData={}".format(self.forwardData))

        self.cols = [KEY_TS, KEY_DATA2, KEY_DATA3] if (config['useCols']['value'] == '') \
                    else config['useCols']['value'].split(',')

        self.enabled = config['enable']['value'] == 'true'
        self.state = 0
        self.last_state = 0

        _LOGGER.setLevel(LOGLEVEL[config['logLevel']['value']])


def plugin_init(config, ingest_ref, callback):
    """ Initialise the plugin.
    Args:
        config: JSON configuration document for the plugin configuration category
        ingest_ref: reference object for ingesting readings
        callback: callback to call to ingest readings to C south service
    Returns:
        data: JSON object to be used in future calls to the plugin
    Raises:
    """

    try:

        config = copy.deepcopy(config)
        _state = CentrifugeState(config, ingest_ref, callback)
        if _state.enabled:

            # create a new process and instantiate an rpc server
            _LOGGER.info("iprpc LOADING module", __name__)
            _module_dir = os.path.dirname(__file__)
            _inference_obj = ModelInferenceHandler()

            _ = _inference_obj.initialize(config)
            config['state'] = _state
            config['inference'] = _inference_obj

    except Exception as ex:
        _LOGGER.exception('Error in initializing ADM_LD_prediction plugin {}'.format(str(ex)))
        raise

    return config


def _clean_filter_input(data, cols):
    """ _clean_filter_input -- fix some common problems in input, stripping excess quotes, 
    Arguments:
        data : [
          dict {
              readings: {
                data_col: data,
                [ts_col: timestamp]
          }
          ...
        ]
    Returns:
        list of dicts of readings data
    """
    if type(data) == dict:
        data = [data]

    if len(data) > 0:
        ts_col = KEY_TS if len(cols) == 0 else cols[0]
        
        has_ts = ts_col in data[0]
        # if you're unlucky enough to have dates enclosed in double quotes, JSON will croak
        for d in data:
            if has_ts:
                d['readings'][ts_col] = d['readings'][ts_col].strip('"')
    
    return data


def plugin_ingest(handle, data):
    """ Returns assets to monitor for centrifuge vibration data
    Args:
        handle: handle returned by the plugin initialisation call
        data: list of
        dicts of
          'readings': {'Channel1': v1, 'Channel2': v2, 'Channel3': v3, 'Channel4': v4, 'user_ts': timestamp},
          'asset': asset-name,
          'id': id,
          'ts': current-timestamp,
          'user_ts': timestamp-of-data
    Action:
        passes data to analysis algorithm, which returns:
        0: LD not required
        1: LD required
        calls the ingest callback to pass along input data
        calls the ingest callback with an asset that holds the current LD required state
    Returns:
    """
    _state = handle['state']
    _inference_obj = handle['inference']
        
    if _state is None or len(data) == 0:
        return  # shut down

    # data = _clean_filter_input(data, _state.cols)

    _LOGGER.debug("ingest: {} {} ".format(len(data), str(data[0])))
    _LOGGER.debug("convert {}".format(str(data[0]['readings'])))

    # pass data up to upper levels
    if _state.forwardData:
        _LOGGER.debug("Forwarding original {} readings".format(len(data)))
        filter_ingest.filter_ingest_callback(_state.callback, _state.ingest_ref, data)

    if not _state.enabled:
        return

    # if incoming reading list has a mix of assets, process only the matching readings & forward non-matching readings
    non_matching_asset_readings = [d for d in data if d['asset'] != _state.ts_asset]
    _LOGGER.debug("Forwarding {} readings with non-matching asset".format(len(non_matching_asset_readings)))
    if not _state.forwardData:
        filter_ingest.filter_ingest_callback(_state.callback, _state.ingest_ref, non_matching_asset_readings)
    del non_matching_asset_readings

    _data = [d for d in data if d['asset'] == _state.ts_asset]
    data = _data
    _LOGGER.debug("Processing {} readings with matching asset".format(len(data)))

    # run model and check need for large discharge
    _result, RE_list = _inference_obj.evaluate(data)
    _LOGGER.debug("plugin_ingest (RPC call) returned: _result={}, type {};  RE_list={}, type={} "
                 .format(_result, type(_result), RE_list, type(RE_list)))
    if len(RE_list) > 1:
        _LOGGER.warn("RE_list has more than 1 element, probably unhandled case")

    if _result >= 0:  # valid
        _state.state = _result

        if _state.state != _state.last_state:
            _LOGGER.info("ingest: state change {}->{} ".format(_state.last_state, _state.state))
            _state.last_state = _state.state

        # must have 1-1 assets we add vs assets we receive
        if 'timestamp' in data[-1]:
            timestamp = data[-1]['timestamp']
        else:
            timestamp = data[-1]['ts']

        asset_dict = {'asset': _state.health_asset,
                      'readings': {'LD_required': _state.state, 'RE_threshold_multiple': float(RE_list[0])},
                      'timestamp': timestamp
                      }

        # pass new asset status data up to upper levels
        _LOGGER.info("send health asset: {}".format([asset_dict]))
        filter_ingest.filter_ingest_callback(_state.callback, _state.ingest_ref, [asset_dict])

    # gc.collect()


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin
    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    ingest_ref = handle['state'].ingest_ref
    callback = handle['state'].callback
    new_handle = plugin_init(new_config, ingest_ref, callback)
    return new_handle


def plugin_shutdown(handle):
    """ Shutdowns the plugin doing required cleanup
    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        plugin shutdown
    """
    _state = handle['state']
    _inference_obj = handle['inference']

    _LOGGER.info('ADM_LD_prediction plugin shut down.')
    
    if _state.enabled:
        _inference_obj.shutdown()

    _state = None
    _inference_obj = None

    handle['state'] = _state
    handle['inference'] = _inference_obj
    _LOGGER.info("ADM_LD_prediction::plugin_shutdown() completed")

